/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#pragma once
#include "exception.h"
#include <string>
#include <cassert>
#include "Math.h"
#include "Platform.h"

namespace SLK
{

//Serializes data into a chunk of memory. The size of the data is the first 2 bytes.
class SerilynExport Serializer
{
public:
	Serializer(unsigned long reserveSize);
	Serializer();
	Serializer(char* data, unsigned long size);
	virtual ~Serializer(void);

	template<typename T>
	Serializer& operator >> (T& val);

	template<typename T>
	Serializer& operator << (const T& val);

	template<class _Elem,class _Traits,class _Ax>
	Serializer& operator >> (std::basic_string<_Elem, _Traits, _Ax>& val);
	
	template<class _Elem,class _Traits,class _Ax>	
	Serializer& operator << (const std::basic_string<_Elem, _Traits, _Ax> &val);

	Serializer& operator << (const char* val);
	Serializer& operator << (const wchar_t* val);

	template<typename T>
	void write(T* data, unsigned long size);

	//Returns the number of bytes written
	template<typename T>
	unsigned long  read(T* buffer, unsigned long bufferSize);

	char* getData();
	unsigned long getDataSize() const;
	void setData(char* data, unsigned long size);

	void compress(int compression = 5);
	void decompress();

	void clear(){m_numBytes=0;};
	void reserve(unsigned long reserveSize, bool zero);

	const Serializer operator=(const Serializer &copy)
	{
		return Serializer(m_data,m_numBytes);
	};
private:

	void checkSize(unsigned long size);

	char* m_data;
	unsigned long m_numBytes;
	unsigned long m_reserveSize;
};

template<typename T>
Serializer& Serializer::operator << ( const T& val)
{
	checkSize(sizeof(T));
	*(T*)(m_data+m_numBytes)=val;	
	m_numBytes+=sizeof(T);

	assert(m_numBytes<=m_reserveSize && "operator<<T");

	return *this;
}

template<typename T>
Serializer& Serializer::operator >> (T& val)
{
	if(m_numBytes-sizeof(T)<0)
		throw Exception("Read past end of serializer");

	val = *(T*)(m_data+m_numBytes-sizeof(T));
	m_numBytes-=sizeof(T);

	assert(m_numBytes>=0 && "operator>>T");

	return *this;
}
	template<class _Elem,class _Traits,class _Ax>
Serializer& Serializer::operator >> (std::basic_string<_Elem, _Traits, _Ax>& val)
{
	short size;
	*this >> size;
	if(m_numBytes-size<0)
		throw Exception("Read past end of serializer");

	val.clear();
	val.resize(size);

	memcpy((void*)val.data(),m_data+m_numBytes-size,size);

	m_numBytes -= size;

	assert(m_numBytes>=0 && "operator>>string");

	return *this;
}
	template<class _Elem,class _Traits,class _Ax>
Serializer& Serializer::operator << (const std::basic_string<_Elem, _Traits, _Ax>& val)
{
	short size = (short)val.size()*sizeof(val[0]);
	checkSize(size);

	for(unsigned long i=0; i<val.size(); i++)
	{
		*this<<val[i];
	}
	*this<<size;

	assert(m_numBytes<=m_reserveSize && "operator<<string");

	return *this;
}
template<typename T>
void Serializer::write(T* data, unsigned long size)
{
	checkSize(sizeof(T)*size + sizeof(size));

	for(unsigned long i=0; i<size; i++)
	{
		*this<<data[size-i-1];
	}
	*this<<size;

	assert(m_numBytes<=m_reserveSize && "write(T* data, unsigned long size)");
}

template<typename T>
unsigned long Serializer::read(T* buffer, unsigned long bufferSize)
{
	unsigned long size;
	*this >> size;

	for(unsigned long i=0; i<size && i<bufferSize; i++)
	{
		*this>>*(buffer++);
	}

	assert(m_numBytes>=0 && "read(T* buffer, unsigned long bufferSize)");

	return min(size,bufferSize);
}


}