/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"
#include "Animator.h"

namespace SLK
{

Animator::Animator( Ogre::Entity *entity ) : 
	m_entity(entity),
	m_target(0),
	m_source(0)
{
}

Animator::Animator( Ogre::Entity *entity, SLK::Serializer &serializer) : 
	m_entity(entity) 
{
	serializer >> m_duration;
	serializer >> m_timeleft;
	serializer >> m_transition;
	serializer >> m_complete;

	m_target = loadAnimationIfExists(serializer);
	m_source = loadAnimationIfExists(serializer);
}

Ogre::AnimationState *Animator::loadAnimationIfExists(SLK::Serializer &serializer)
{
	bool hasAnim;
	Ogre::AnimationState *anim= 0;
	serializer >> hasAnim;
	if(hasAnim)
	{
		bool enabled;
		serializer >> enabled;
		bool loop;
		serializer >> loop;
		Ogre::Real weight;
		serializer >> weight;
		Ogre::Real time;
		serializer >> time;
		Ogre::String name;
		serializer >> name;
		anim = m_entity->getAnimationState(name);
		anim->setEnabled(enabled);
		anim->setLoop(loop);
		anim->setWeight(weight);
		anim->setTimePosition(time);
	}

	return anim;
}

void Animator::save(SLK::Serializer &serializer)
{
	saveAnimation(serializer, m_source);
	saveAnimation(serializer, m_target);

	serializer << m_complete;
	serializer << m_transition;
	serializer << m_timeleft;
	serializer << m_duration;
}

void Animator::saveAnimation(SLK::Serializer &serializer, Ogre::AnimationState *anim)
{
	if(anim)
	{
		serializer << anim->getAnimationName();
		serializer << anim->getTimePosition();
		serializer << anim->getWeight();
		serializer << anim->getLoop();
		serializer << anim->getEnabled();
		serializer << true;
	}
	else
	{
		serializer << false;
	}
}


void Animator::init(const Ogre::String &animation, bool loop)
{
	assert(m_entity && "null entity in animator");

	Ogre::AnimationStateSet *set = m_entity->getAllAnimationStates();
	if(!set)
		throw SLK::Exception("Entity %s has no animations", m_entity->getName());
	Ogre::AnimationStateIterator it = set->getAnimationStateIterator();
	while(it.hasMoreElements())
	{
		Ogre::AnimationState *anim = it.getNext();
		anim->setEnabled(false);
		anim->setWeight(0);
		anim->setTimePosition(0);
	}
	m_source = m_entity->getAnimationState( animation );
	m_source->setEnabled(true);
	m_source->setWeight(1);
	m_source->setLoop(loop);
	m_timeleft = 0;
	m_duration = 1;
	m_target = 0;
	m_complete = false;
} 

void Animator::blend( const Ogre::String &animation, BlendingTransition transition, Ogre::Real duration, bool loop )
{
  if( transition == BlendSwitch )
  {
     if( m_source != 0 )
        m_source->setEnabled(false);
     m_source = m_entity->getAnimationState( animation );
     m_source->setEnabled(true);
     m_source->setWeight(1);
     m_source->setTimePosition(0);
	 m_source->setLoop(loop);

     m_timeleft = 0;
  } 
  else 
  { 
     Ogre::AnimationState *newTarget = m_entity->getAnimationState( animation );
	 newTarget->setLoop(loop);

     if( m_timeleft > 0 )
     {
        // oops, weren't finished yet
        if( newTarget == m_target )
        {
           // nothing to do! (ignoring duration here)
        }
        else if( newTarget == m_source )
        {
           // going back to the source state, so let's switch
           m_source = m_target;
           m_target = newTarget;
           m_timeleft = m_duration - m_timeleft; // i'm ignoring the new duration here
        }
        else
        {
           // ok, newTarget is really new, so either we simply replace the target with this one, or
           // we make the target the new source
           if( m_timeleft < m_duration * 0.5 )
           {
              // simply replace the target with this one
              m_target->setEnabled(false);
              m_target->setWeight(0);
           }
           else
           {
              // old target becomes new source
              m_source->setEnabled(false);
              m_source->setWeight(0);
              m_source = m_target;

           } 
           m_target = newTarget;
           m_target->setEnabled(true);
           m_target->setWeight( 1.0 - m_timeleft / m_duration );
           m_target->setTimePosition(0);
        }
     }
     else
     {
        if( newTarget == m_source && m_target == 0 )
		{
			//Nothing to do
		}
		else
		{
			m_transition = transition;
			m_timeleft = m_duration = duration;
			m_target = newTarget;
			m_target->setEnabled(true);
			m_target->setWeight(0);
			m_target->setTimePosition(0);
		}
     }
  }
}


void Animator::update( Ogre::Real time )
{
  if( m_source != 0 )
  {
     if( m_timeleft > 0 )
     {
        m_timeleft -= time;

        if( m_timeleft <= 0 )
        {
           // finish blending
           m_source->setEnabled(false);
           m_source->setWeight(0);
           m_source = m_target;
           m_source->setEnabled(true);
           m_source->setWeight(1);
           m_target = 0;
        }
        else
        {
           // still blending, advance weights
           m_source->setWeight(m_timeleft / m_duration);
           m_target->setWeight(1.0 - m_timeleft / m_duration);

           if(m_transition == Animator::BlendWhileAnimating)
              m_target->addTime(time);
        }
     }

     if (m_source->getTimePosition() >= m_source->getLength())
     {
        m_complete=true;
     }
     else
     {
       m_complete=false;
     }

     m_source->addTime(time);
  }
}


}