/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"
#include "Application.h"
#include "Exception.h"

namespace SLK
{
Application* Application::getSingletonPtr()
{
	return m_singleton;
}

Application& Application::getSingleton()
{
	assert(m_singleton); 
	return (*m_singleton);
}

Application::Application(void):
	m_ogreRoot(0),
	m_window(0),
	m_state(0),
	m_viewport(0),
	m_camera(0),
	m_sceneManager(0),
	m_soundManager(0),
	m_input(0),
	m_scripting(0)
#ifdef OGRE_STATIC_LIB
	,m_glPlugin(0),
	m_d3d9Plugin(0),
	m_d3d10Plugin(0),
	m_d3d11Plugin(0),
	m_octreePlugin(0),
	m_cgPlugin(0),
	m_particleFXPlugin(0)
#endif
#ifdef USE_PARTICLE_UNIVERSE
	,m_particleUniversePlugin(0)
#endif
{
}

Application::~Application(void)
{
	SAFEDELETE(m_input);
	SAFEDELETE(m_soundManager);
	SAFEDELETE(m_scripting);

     m_sceneManager->destroyCamera(m_camera);

	 DestroyAllChildrenOfNode(m_sceneManager->getRootSceneNode());

	if(m_sceneManager)
		m_ogreRoot->destroySceneManager(m_sceneManager);

	SAFEDELETE(m_ogreRoot);

#ifdef OGRE_STATIC_LIB
	SAFEDELETE(m_glPlugin);
	SAFEDELETE(m_d3d9Plugin);
	SAFEDELETE(m_d3d10Plugin);
	SAFEDELETE(m_d3d11Plugin);
	SAFEDELETE(m_octreePlugin);
	SAFEDELETE(m_cgPlugin);
	SAFEDELETE(m_particleFXPlugin);
#endif
#ifdef USE_PARTICLE_UNIVERSE
	SAFEDELETE(m_particleUniversePlugin);
#endif


}

#ifdef _DEBUG
#define pluginSuffix std::string("_d")
#else
	#define pluginSuffix std::string("")
#endif

bool Application::init(unsigned int flags, State *initialState, std::string title, std::vector<ResourceInitializer> &resources, bool quitOnESC)
{
	m_ogreRoot = new Ogre::Root("");

	m_libFlags = flags;
	if(flags & LibFlags::OgreRendererOpenGL)
	{
#ifdef OGRE_STATIC_LIB
		m_glPlugin = new Ogre::GLPlugin();
		m_ogreRoot->installPlugin(m_glPlugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/RenderSystem_GL") + pluginSuffix);
#endif
	}
	if(flags & LibFlags::OgreRendererDX9)
	{
#ifdef OGRE_STATIC_LIB
		m_d3d9Plugin = new Ogre::D3D9Plugin();
		m_ogreRoot->installPlugin(m_d3d9Plugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/RenderSystem_Direct3D9") + pluginSuffix);
#endif
	}
	if(flags & LibFlags::OgreRendererDX10)
	{
#ifdef OGRE_STATIC_LIB
		m_d3d10Plugin = new Ogre::D3D10Plugin();
		m_ogreRoot->installPlugin(m_d3d10Plugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/RenderSystem_Direct3D10") + pluginSuffix);
#endif
	}
	if(flags & LibFlags::OgreRendererDX11)
	{
#ifdef OGRE_STATIC_LIB
		m_d3d11Plugin = new Ogre::D3D11Plugin();
		m_ogreRoot->installPlugin(m_d3d11Plugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/RenderSystem_Direct3D11") + pluginSuffix);
#endif
	}
	if(!(flags & (LibFlags::OgreRendererOpenGL | LibFlags::OgreRendererDX9 | LibFlags::OgreRendererDX10 | LibFlags::OgreRendererDX11)))
		throw SLK::Exception("Flags must include at least one renderer");
	if(flags & LibFlags::OgreOctreeSceneManager)
	{
#ifdef OGRE_STATIC_LIB
		m_octreePlugin = new Ogre::OctreePlugin();
		m_ogreRoot->installPlugin(m_octreePlugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/Plugin_OctreeSceneManager") + pluginSuffix);
#endif
	}
	else
	{
		throw SLK::Exception("Flags must include at least one Scene Manager");
	}

	if(flags & LibFlags::OgreCGShaders)
	{
#ifdef OGRE_STATIC_LIB
		m_cgPlugin = new Ogre::CgPlugin();
		m_ogreRoot->installPlugin(m_cgPlugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/Plugin_CgProgramManager") + pluginSuffix);
#endif
	}
#ifdef USE_PARTICLE_UNIVERSE
	if(flags & LibFlags::ParticleUniverse)
	{
		m_particleUniversePlugin = new ParticleUniverse::ParticleUniversePlugin();
		m_ogreRoot->installPlugin(m_particleUniversePlugin);

		if(flags & LibFlags::OgreParticleFX)
		{
			Ogre::LogManager::getSingletonPtr()->logMessage("Particle Universe taking precidence over Ogre ParticleFX. Both cannot be used at the same time.",Ogre::LML_TRIVIAL);
		}
	}
	else 
#endif
	if(flags & LibFlags::OgreParticleFX)
	{
#ifdef OGRE_STATIC_LIB
		m_particleFXPlugin = new Ogre::ParticleFXPlugin();
		m_ogreRoot->installPlugin(m_particleFXPlugin);
#else
		m_ogreRoot->loadPlugin(std::string("Plugins/Plugin_ParticleFX") + pluginSuffix);
#endif
	}

	//Initialize Ogre
    if(m_ogreRoot->showConfigDialog())
    {
        m_window = m_ogreRoot->initialise(true,title);
		m_window->setDeactivateOnFocusChange(false);
    }
    else
    {
        return false;
    }

	m_sceneManager = m_ogreRoot->createSceneManager(Ogre::ST_GENERIC);

	// Initialize sound if needed
	if(flags & LibFlags::OgreOggSound)
	{
		try
		{
			m_ogreRoot->loadPlugin(std::string("OgreOggSound") + pluginSuffix);

			m_soundManager = new SoundManager();
		}
		catch(Ogre::Exception &)
		{
			Ogre::LogManager::getSingletonPtr()->logMessage("Please run the OpenAL installer");
			exit(1);
		}
		catch(std::exception)
		{
			Ogre::LogManager::getSingletonPtr()->logMessage("Please run the OpenAL installer");
			exit(1);
		}
		catch(...)
		{
			Ogre::LogManager::getSingletonPtr()->logMessage("Please run the OpenAL installer");
			exit(1);
		}
	}
	// Initialize scripting if needed
	if(flags & LibFlags::AngelScript)
	{
		m_scripting = new Scripting();
	}

    m_camera = m_sceneManager->createCamera("Camera");

    // Look straight down
    m_viewport = m_window->addViewport(m_camera);
	m_viewport->setBackgroundColour(Ogre::ColourValue(0,0,0));

	m_camera->setAspectRatio(Ogre::Real(m_viewport->getActualWidth()) / Ogre::Real(m_viewport->getActualHeight()));

	for(unsigned int i=0; i<resources.size(); i++)
	{
		Ogre::ResourceGroupManager::getSingleton().addResourceLocation(
				resources[i].location, resources[i].type, resources[i].group);
	}

	Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();

	m_ogreRoot->renderOneFrame();

	m_input = new UIManager(m_window, m_sceneManager);
	m_input->init();

	if(quitOnESC)
		m_input->getHotkeyManager()->assignHotkey(&Application::KeyboardDown,this, OIS::KC_ESCAPE);

	m_state = initialState;
	m_state->init();

	m_running = true;

	return true;
}

bool Application::run()
{
	Ogre::WindowEventUtilities::messagePump();
	m_ogreRoot->renderOneFrame();

	m_timer.update();

	float deltaTime = m_timer.getDeltaTime();
	m_input->update(deltaTime);

	if(m_libFlags & LibFlags::OgreOggSound)
	{
		m_soundManager->update(deltaTime);
	}

	if(!m_state->update(deltaTime))
	{
		State *oldState = m_state;
		m_state = m_state->next();
		if(oldState != m_state)
		{
			if(m_state)
				m_state->init();
			SAFEDELETE(oldState);
		}
	}

	bool running = m_state && !m_window->isClosed();
	running = running && m_running;

	if(!running)
		SAFEDELETE(m_state);
	return running;
}

bool Application::KeyboardDown(SLK::EventArgs &args)
{
	m_running = false;

	return true;
}


void Application::DestroyAllChildrenOfNode(Ogre::SceneNode *node)
{
   if ( !node )
   {
      assert( false && "destroying null node");
      return;
   }

   // Destroy all the attached objects
   Ogre::SceneNode::ObjectIterator itObject = node->getAttachedObjectIterator();

   while ( itObject.hasMoreElements() )
   {
      Ogre::MovableObject* pObject = static_cast<Ogre::MovableObject*>(itObject.getNext());
	  std::stringstream str;
	  str << "**Destroying object dropping " << pObject->getName();
	  Ogre::LogManager::getSingletonPtr()->logMessage(str.str());

      node->getCreator()->destroyMovableObject( pObject );
   }

   // Recurse to child SceneNodes
   Ogre::SceneNode::ChildNodeIterator itChild = node->getChildIterator();

   while ( itChild.hasMoreElements() )
   {
      Ogre::SceneNode* pChildNode = static_cast<Ogre::SceneNode*>(itChild.getNext());
	  std::stringstream str;
	  str << "**Destroying sceneNode dropping " << pChildNode->getName();
	  Ogre::LogManager::getSingletonPtr()->logMessage(str.str());

      DestroyAllChildrenOfNode( pChildNode );
   }

}

}