/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"
#include "CameraController3rdPerson.h"
#include "Application.h"
#include "Serializer.h"
#include "DBStatement.h"
#include "DBResultSet.h"

namespace SLK
{

CameraController3rdPerson::CameraController3rdPerson(Ogre::Camera *camera, Ogre::SceneNode *parentNode, Ogre::Degree yaw, Ogre::Degree pitch, float zoom):
	m_camera(camera),
	m_pitch(pitch),
	m_yaw(yaw),
	m_parentNode(parentNode),
	m_moveForward(false),
	m_moveBackward(false),
	m_moveLeft(false),
	m_moveRight(false)
{
	m_cameraRotate = parentNode->createChildSceneNode();
	m_cameraNode = m_cameraRotate->createChildSceneNode();

	m_cameraNode->attachObject(m_camera);

	//setYaw(yaw);
	setPitch(pitch);

	moveZoom(zoom,0,zoom+1);
}

CameraController3rdPerson::~CameraController3rdPerson(void)
{
	m_cameraNode->detachObject(m_camera);
}

void CameraController3rdPerson::save(SLK::Database *database)
{
	SLK::Serializer serializer;
	Ogre::Vector3 pos =m_parentNode->getPosition();
	serializer << pos.x;
	serializer << pos.y;
	serializer << pos.z;

	pos =m_cameraNode->getPosition();
	serializer << pos.x;
	serializer << pos.y;
	serializer << pos.z;

	Ogre::Quaternion rot = m_cameraRotate->getOrientation();
	serializer << rot.x;
	serializer << rot.y;
	serializer << rot.z;
	serializer << rot.w;

	SLK::DBStatement statement(database);
	static std::string MAIN_DEF =
	"UPDATE 'Main' SET Camera=@Camera;";
	statement.prepare(MAIN_DEF);
	statement.bind("@Camera", serializer);
	statement.run();

}

void CameraController3rdPerson::load(SLK::Database *database)
{
	static std::string MAIN_DEF = "SELECT Camera FROM 'Main'";
	SLK::DBResultSet results;
	SLK::DBStatement statement(database);
	statement.prepare(MAIN_DEF);

	statement.run(&results);
	statement.reset();

	SLK::Serializer *serializer = 0;
	results.GetRow(0)->Get("Camera",serializer);
	if(serializer->getDataSize()>0)
	{
		Ogre::Quaternion rot;
		(*serializer) >> rot.w;
		(*serializer) >> rot.z;
		(*serializer) >> rot.y;
		(*serializer) >> rot.x;
		m_cameraRotate->setOrientation(rot);

		Ogre::Vector3 pos;
		(*serializer) >> pos.z;
		(*serializer) >> pos.y;
		(*serializer) >> pos.x;
		m_cameraNode->setPosition(pos);

		(*serializer) >> pos.z;
		(*serializer) >> pos.y;
		(*serializer) >> pos.x;
		m_parentNode->setPosition(pos);
	}

	SAFEDELETE(serializer);
}


void CameraController3rdPerson::updateMove(float distance)
{
	Ogre::Vector3 translate(0,0,0);
	if(m_moveForward)
		translate.z += -1;
	if(m_moveBackward)
		translate.z += 1;
	if(m_moveLeft)
		translate.x += -1;
	if(m_moveRight)
		translate.x += 1;

	Ogre::Quaternion orientation = m_cameraRotate->getOrientation();
	Ogre::Vector3 finalTranslate = orientation * translate * distance;
	finalTranslate.y = 0; //Stick the player to the ground
	m_parentNode->translate(finalTranslate);
}

void CameraController3rdPerson::addPitch(Ogre::Degree pitch, Ogre::Degree minPitch, Ogre::Degree maxPitch)
{
	m_pitch += pitch;

	m_pitch = std::max(m_pitch,minPitch);
	m_pitch = std::min(m_pitch,maxPitch);
	
	setPitch(m_pitch);
}

void CameraController3rdPerson::addYaw(Ogre::Degree yaw, Ogre::Degree minYaw, Ogre::Degree maxYaw)
{
	m_yaw += yaw;
	m_yaw = std::max(m_yaw,minYaw);
	m_yaw = std::min(m_yaw,maxYaw);

	setYaw(m_yaw);
}


void CameraController3rdPerson::addPitch(Ogre::Degree pitch)
{
	m_pitch += pitch;
	
	setPitch(m_pitch);
}

void CameraController3rdPerson::addYaw(Ogre::Degree yaw)
{
	m_yaw += yaw;

	setYaw(m_yaw);
}

void CameraController3rdPerson::moveZoom(float zoomLevel, float minZoom, float maxZoom)
{
	m_cameraNode->translate(0,0,zoomLevel);
	Ogre::Vector3 pos = m_cameraNode->getPosition();
	if(pos.z > maxZoom)
		m_cameraNode->setPosition(0,0,maxZoom);
	else if(pos.z < minZoom)
		m_cameraNode->setPosition(0,0,minZoom);
}

void CameraController3rdPerson::setYaw(Ogre::Degree yaw)
{
	m_yaw = yaw;

	m_cameraRotate->setOrientation(Ogre::Quaternion::IDENTITY);
	m_cameraRotate->yaw(Ogre::Radian(m_yaw),Ogre::Node::TS_LOCAL );
	m_cameraRotate->pitch(Ogre::Radian(m_pitch),Ogre::Node::TS_LOCAL);
}

void CameraController3rdPerson::setPitch(Ogre::Degree pitch)
{
	m_pitch = pitch;

	m_cameraRotate->setOrientation(Ogre::Quaternion::IDENTITY);
	m_cameraRotate->pitch(Ogre::Radian(m_pitch),Ogre::Node::TS_LOCAL);
	m_cameraRotate->yaw(Ogre::Radian(m_yaw),Ogre::Node::TS_LOCAL);
}

}