/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"


#include "Database.h"
#include "DBResultSet.h"
#include "DBStatement.h"

namespace SLK
{

Database::Database(void):
	m_database(0)
{
}

Database::~Database(void)
{
	close();

	for(std::map<std::string, Sequencer*>::iterator iter = m_sequencers.begin(); iter!=m_sequencers.end(); iter++)
	{
		SAFEDELETE(iter->second);
	}
}

Sequencer *Database::getSequencer(std::string tableName, std::string idRow)
{
	std::map<std::string, Sequencer*>::iterator iter = m_sequencers.find(tableName);
	if(iter !=m_sequencers.end())
		return iter->second;
	Sequencer *sequencer = new Sequencer(tableName, idRow, this);
	m_sequencers.insert(std::make_pair(tableName, sequencer));
	return sequencer;
}

void Database::open(std::string fileName)
{
	assert(!m_database && "Database already open");
	
	boost::filesystem::path path(getSaveFolder());
	if(!boost::filesystem::exists(path))
	{
		boost::filesystem::create_directory(path);
	}
	path = path / fileName;
	if(!boost::filesystem::exists(path))
	{
		sqlite3_open_v2(path.string().c_str(),&m_database, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX,0);
	}
	else
	{
		sqlite3_open_v2(path.string().c_str(),&m_database, SQLITE_OPEN_READWRITE | SQLITE_OPEN_FULLMUTEX,0);
	}

}

void Database::close()
{
	if(m_database)
	{
		// Close open statements
		sqlite3_stmt *pStmt;
		while( (pStmt = sqlite3_next_stmt(m_database, 0))!=0 )
		{
			 sqlite3_finalize(pStmt);
		} 

		int result = sqlite3_close(m_database);
		assert(result == SQLITE_OK && "Error closing database");
		m_database = 0;
	}
}

bool Database::tableExists(std::string tableName)
{
	if(!m_database)
		return false;

	static std::string FINDTABLES_DEF = 
	"SELECT name FROM sqlite_master WHERE name=@TableName";

	DBResultSet results;
	DBStatement statement(m_database);
	statement.prepare(FINDTABLES_DEF);

	statement.bind("@TableName",tableName);

	statement.run(&results);
	statement.reset();
	return results.GetRowCount() > 0;
}

std::string Database::getSaveFolder()
{
	// read these from the ubiquitous config file...on Win32 we have a nice handy
	// API to read config files; on other platforms we'll need to fake one
	TCHAR path[3*MAX_PATH+1];

#ifdef WIN32
	int status = SHGetFolderPath(NULL, CSIDL_MYDOCUMENTS, NULL, SHGFP_TYPE_CURRENT, path);
	assert(status == S_OK && "Bad my docs folder?");
#else
	// Linux and Mac go here
#endif
	std::string pathname(path);

#ifdef _UNICODE
	pathname += L"/" + Utility::toWideString(m_appName) + L"/";
#else
	pathname += "/" + std::string("GoblinHome") + "/";
#endif
	
	return pathname;
}

}
