/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"


#include "Script.h"
#include "Scripting.h"
#include "DebugConsole.h"

namespace SLK
{

Script::Script(std::string &fileName):
	m_loadOK(true),
	m_scriptContext(0)
{
	std::string fullpath = "Media/Scripts/";
	fullpath += fileName;

	// TODO: This should only build if the particular file has not been built yet. Need some sort of manager.
	m_scriptModule = Scripting::getSingleton().getEngine()->GetModule(fileName.c_str(), asGM_CREATE_IF_NOT_EXISTS);
	loadScript(fullpath.c_str());
	if(m_loadOK)
	{
		int r = m_scriptModule->Build();
		if(r<0)
		{
			m_loadOK=false;
			Ogre::LogManager::getSingletonPtr()->logMessage("Build Failed.");
			//DebugConsole::getSingleton().addText("Build Failed.");
		}
		else
		{
			m_scriptContext = Scripting::getSingleton().getEngine()->CreateContext();
		}
	}
}

Script::Script(std::string &id, std::string scriptString):
	m_loadOK(true),
	m_scriptContext(0)
{
	// TODO: This should only build if the particular file has not been built yet. Need some sort of manager.
	m_scriptModule = Scripting::getSingleton().getEngine()->GetModule(id.c_str(), asGM_CREATE_IF_NOT_EXISTS);

	int r = m_scriptModule->AddScriptSection(id.c_str(), scriptString.c_str());
	if(r<0)
	{
		m_loadOK = false;
	}
	else
	{
		r = m_scriptModule->Build();
		if(r<0)
		{
			m_loadOK=false;
			Ogre::LogManager::getSingletonPtr()->logMessage("Build Failed.");
			//DebugConsole::getSingleton().addText("Build Failed.");
		}
		else
		{
			m_scriptContext = Scripting::getSingleton().getEngine()->CreateContext();
		}
	}

}

Script::~Script(void)
{
	if(m_scriptContext)
		m_scriptContext->Release();
}

void Script::prepareAndRun()
{
	prepare();
	run();
}

bool Script::prepare(std::string signature)
{
	if(m_scriptContext)
	{
		int funcId  = m_scriptModule->GetFunctionIdByDecl(signature.c_str());
		if(funcId<0)
		{
			Ogre::LogManager::getSingletonPtr()->logMessage(std::string("Function with signature \"") + signature + "\" not found.");
			return false;
		}

		m_scriptContext->Prepare(funcId);
		return true;
	}
	return false;
}

void *Script::run()
{
	if(m_scriptContext)
	{
		int returnVal = m_scriptContext->Execute();
		if(returnVal == asEXECUTION_EXCEPTION)
		{
			asIScriptEngine *engine = Scripting::getSingleton().getEngine();

			// Determine the exception that occurred
			char buffer[1024];
			sprintf(buffer,"desc: %s\n", m_scriptContext->GetExceptionString());
			DebugConsole::getSingleton().addText(buffer);

			// Determine the function where the exception occurred
			int funcId = m_scriptContext->GetExceptionFunction();
			const asIScriptFunction *function = engine->GetFunctionDescriptorById(funcId);
			sprintf(buffer,"func: %s\n", function->GetDeclaration());
			DebugConsole::getSingleton().addText(buffer);
			sprintf(buffer,"modl: %s\n", function->GetModuleName());
			DebugConsole::getSingleton().addText(buffer);
			sprintf(buffer,"sect: %s\n", function->GetScriptSectionName());
			DebugConsole::getSingleton().addText(buffer);

			// Determine the line number where the exception occurred
			sprintf(buffer,"line: %d\n", m_scriptContext->GetExceptionLineNumber());
			DebugConsole::getSingleton().addText(buffer);

			return 0;
		}
	}
	return 0;
}

//taken from "include" sample of angelscript
void Script::loadScript(const char *file, std::set<std::string> *includedScripts)
{

	bool ownsSet = false;

	if(!includedScripts)
	{
		includedScripts = new std::set<std::string>();
		ownsSet = true;
	}

	// TODO: The file name stored in the set should be the fully resolved name because
	// it is possible to name the same file in multiple ways using relative paths.

	// We will load the script from a file on the disk.
	FILE *f = fopen(file, "rb");
	if( f == 0 )
	{
		Ogre::LogManager::getSingletonPtr()->logMessage(std::string("Failed to open the script file.") + std::string(file));
		//DebugConsole::getSingleton().addText("Failed to open the script file.");
		if(ownsSet)
			SAFEDELETE(includedScripts);	
		m_loadOK = false;
		return;
	}

	// Has the script been loaded already?
	std::string scriptFile = file;
	if( includedScripts->find(scriptFile) != includedScripts->end() )
	{
		// Already loaded 
		//cout << "File '" << scriptFile << "' already included. Ignoring." << endl;
		if(ownsSet)
			SAFEDELETE(includedScripts);
		return;
	}

	// Included the script as loaded
	//cout << "Loading '" << scriptFile << "'." << endl;
	includedScripts->insert(scriptFile);

	// Determine the size of the file	
	fseek(f, 0, SEEK_END);
	int len = ftell(f);
	fseek(f, 0, SEEK_SET);

	// On Win32 it is possible to do the following instead
	// int len = _filelength(_fileno(f));

	// Read the entire file
	std::string script;
	script.resize(len);
	int c =	fread(&script[0], len, 1, f);
	fclose(f);

	if( c == 0 ) 
	{
		//cout << "Failed to load script file." << endl;
		if(ownsSet)
			SAFEDELETE(includedScripts);
		m_loadOK = false;
		return;
	}

	// Search the script file for #include directives
	int r = processInclude(script,includedScripts);
	if( r < 0 ) 
	{
		if(ownsSet)
			SAFEDELETE(includedScripts);
		return;
	}

	// Add the script sections that will be compiled into executable code.
	// If we want to combine more than one file into the same script, then 
	// we can call AddScriptSection() several times for the same module and
	// the script engine will treat them all as if they were one. The script
	// section name, will allow us to localize any errors in the script code.

	// We'll tell AngelScript to make a copy of the script buffer, as we won't keep it in memory (last param).
	r = m_scriptModule->AddScriptSection(file, &script[0]);
	if( r < 0 ) 
	{
		if(ownsSet)
			SAFEDELETE(includedScripts);
		m_loadOK = false;
		//cout << "AddScriptSection() failed" << endl;
		return;
	}

	SAFEDELETE(includedScripts);

}

//taken from "include" sample of angelscript
int Script::processInclude(std::string &script, std::set<std::string> *includedScripts)
{
	// Go through the script line by line, searching for #include directives
	for( int pos = 0; pos != script.npos && pos < (signed)script.length(); )
	{
		// Find the end of the line
		int end = script.find('\n', pos);
		if( end == script.npos ) 
		{
			// End of script
			end = script.length(); 
		}
		else 
		{
			// Don't include the newline character
			--end; 
		}
		
		// Skip white spaces
		int start = script.find_first_not_of(" \t", pos, 2);
		if( start == script.npos || start > end ) start = end;
		
		if( start < end && script[start] == '#' && (end-start) > 10 )
		{
			// Is it an #include directive?
			if( strncmp("#include", &script[start], 8) == 0 )
			{
				start += 8;
				
				// Skip white spaces
				start = script.find_first_not_of(" \t", start, 2);
				if( start == script.npos || start > end ) start = end;
				
				if( script[start] == '\"' )
				{
					++start;
					
					// Find second "
					int stop = script.find('\"', start);
					if( stop != script.npos && stop <= end )
					{
						std::string filename = script.substr(start, stop-start);
						
						// Load the included script (if not already loaded)
						loadScript(filename.c_str());
						
						// Comment this line so that the compiler doesn't complain
						script[pos] = '/';
						script[pos+1] = '/';
					}
				}
			}
		}
		
		// Move to the next line
		pos = end + 2;
	}

	return 0;
}

}
