/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"

#include "Serializer.h"

namespace SLK
{

Serializer::Serializer(void):
	m_data(0),
	m_numBytes(0),
	m_reserveSize(10)
{
	m_data = new char[m_reserveSize];
}

Serializer::Serializer(unsigned long reserveSize):
	m_data(0),
	m_numBytes(0),
	m_reserveSize(reserveSize)
{
	m_data = new char[reserveSize];
}

Serializer::Serializer(char* data, unsigned long size):
	m_data(0),
	m_numBytes(0),
	m_reserveSize(size)
{
	m_data = new char[size];
	checkSize(size);
	memcpy(m_data,data,size);
	m_numBytes = size;
}


Serializer::~Serializer(void)
{
	SAFEDELETEARRAY(m_data);
}



char* Serializer::getData()
{
	return m_data;
}

unsigned long Serializer::getDataSize() const
{
	return m_numBytes;
}

void Serializer::setData(char* data, unsigned long size)
{
	checkSize(size);
	SAFEDELETEARRAY(m_data);
	m_data = data;
	m_numBytes = size;
}

void Serializer::checkSize(unsigned long size)
{
	while(m_numBytes+size>m_reserveSize)
	{
		unsigned long newSize = (unsigned long)(m_reserveSize*1.5);
		if(newSize<m_numBytes+size)
			newSize = m_numBytes+size;
		char *temp = new char[newSize];
		memcpy(temp, m_data, m_reserveSize);
		m_reserveSize = newSize;
		SAFEDELETEARRAY(m_data);
		m_data = temp;
	}
}

void Serializer::reserve(unsigned long reserveSize, bool zero)
{
	checkSize(reserveSize);
	if(zero)
		memset(m_data,0,m_reserveSize);
}

Serializer& Serializer::operator << (const char* val)
{
	short size=0;
	while(*val !=0)
	{
		checkSize(1);
		*this<<(*val);
		val++;
		size++;

	}
	*this<<size;

	assert(m_numBytes<=m_reserveSize && "operator<<char*");

	return *this;
}

Serializer& Serializer::operator << (const wchar_t* val)
{
	short size=0;
	short sizeOfOne = sizeof(wchar_t);
	while(*val !=0)
	{
		checkSize(sizeOfOne);
		*this<<(*val);
		size+=sizeOfOne;
		val++;

	}
	*this<<size;

	assert(m_numBytes<=m_reserveSize && "operator<<wchar*");

	return *this;
}

void Serializer::compress(int compression)
{
	unsigned long destLen = compressBound(m_numBytes);

	char* dest = new char[destLen];
	int result = compress2((Bytef*)dest,(uLongf*)&destLen,(const Bytef*)m_data,(uLongf)m_numBytes,compression);
	if(result != Z_OK)
		throw SLK::Exception("Serializer::compress returned zlib error %d",result);

	unsigned long oldLen = m_numBytes;
	clear();

	memcpy((void*)m_data,dest,destLen);
	m_numBytes = destLen;
	if(m_reserveSize < destLen)
	{
		SAFEDELETEARRAY(m_data);
		m_reserveSize = destLen;
		m_data = new char[m_reserveSize];
	}
	*this << oldLen;

	SAFEDELETEARRAY(dest);
}

void Serializer::decompress()
{
	char *dest;
	unsigned long destLen;
	*this >> destLen;
	dest = new char[destLen];

	uncompress((Bytef*)dest,&destLen,(const Bytef*)m_data,m_numBytes);

	clear();
	if(m_reserveSize < destLen)
	{
		SAFEDELETEARRAY(m_data);
		m_reserveSize = destLen;
		m_data = new char[m_reserveSize];
	}
	memcpy((void*)m_data,dest,destLen);
	m_numBytes = destLen;

	SAFEDELETEARRAY(dest);
}


}