/*
-----------------------------------------------------------------------------
Copyright (c) 2009-2010 Sara Kempen

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "StdAfx.h"

#include "UIManager.h"
#include "Event.h"

namespace SLK
{

UIManager::UIManager(Ogre::RenderWindow *window, Ogre::SceneManager *sceneManager):
	m_window(window),
	m_sceneManager(sceneManager),
	m_keyboard(0),
	m_mouse(0),
	m_inputManager(0),
	m_hotkeyManager(0),
	m_guiSystem(0)
{

}

UIManager::~UIManager(void)
{
    if( m_inputManager ) 
	{
        if( m_mouse ) 
		{
            m_inputManager->destroyInputObject( m_mouse );
            m_mouse = 0;
        }

        if( m_keyboard ) 
		{
            m_inputManager->destroyInputObject( m_keyboard );
            m_keyboard = 0;
        }
/*
        if( mJoysticks.size() > 0 ) {
            itJoystick    = mJoysticks.begin();
            itJoystickEnd = mJoysticks.end();
            for(; itJoystick != itJoystickEnd; ++itJoystick ) {
                mInputSystem->destroyInputObject( *itJoystick );
            }

            mJoysticks.clear();
        }*/

        m_inputManager->destroyInputSystem( m_inputManager );
        m_inputManager = 0;
    }
	_CrtCheckMemory();

	CEGUI::OgreRenderer::destroySystem();

	SAFEDELETE(m_hotkeyManager);

}

bool UIManager::init()
{


	size_t windowHnd = 0;
	std::ostringstream windowHndStr;
	OIS::ParamList pl;

	m_window->getCustomAttribute("WINDOW", &windowHnd);
	windowHndStr << windowHnd;
	pl.insert(std::make_pair(std::string("WINDOW"), windowHndStr.str()));
	m_inputManager = OIS::InputManager::createInputSystem(pl);

	try
	{

        m_keyboard = static_cast<OIS::Keyboard*>( m_inputManager->createInputObject( OIS::OISKeyboard, true ) );
        m_keyboard->setEventCallback( this );


        m_mouse = static_cast<OIS::Mouse*>( m_inputManager->createInputObject( OIS::OISMouse, true ) );
        m_mouse->setEventCallback( this );

        // Get window size
        unsigned int width, height, depth;
        int left, top;
        m_window->getMetrics( width, height, depth, left, top );

        // Set mouse region
        this->setWindowExtents( width, height );

		m_hotkeyManager = new HotkeyManager(m_keyboard);

	}
	catch (const OIS::Exception &e)
	{
		throw Ogre::Exception(42, e.eText, "Application::setupInputSystem");
	}

   // CEGUI setup
	m_guiRenderer = &CEGUI::OgreRenderer::bootstrapSystem(*m_window);// new CEGUI::OgreRenderer(m_window, Ogre::RENDER_QUEUE_OVERLAY, false, 3000, m_sceneManager);
	m_guiSystem = CEGUI::System::getSingletonPtr();//new CEGUI::System(m_guiRenderer);
	CEGUI::Logger::getSingleton().setLoggingLevel(CEGUI::Errors);

	// load in the scheme file, which auto-loads the WindowsLook and WindowsLook imageset
	CEGUI::SchemeManager::getSingleton().create( "WindowsLook.scheme" );

	m_guiSystem->setDefaultMouseCursor((CEGUI::utf8*)"WindowsLook", (CEGUI::utf8*)"MouseArrow");
	CEGUI::FontManager::getSingleton().create("Vera.font");
	CEGUI::FontManager::getSingleton().create("VeraMono.font");
	CEGUI::FontManager::getSingleton().create("VeraBold.font");

	CEGUI::System::getSingleton().setDefaultTooltip("WindowsLook/Tooltip");

	m_guiSystem->setDefaultFont((CEGUI::utf8*)"Vera");

	// set the mouse cursor initially in the middle of the screen
	m_guiSystem->injectMousePosition((float)m_window->getWidth() / 2.0f, (float)m_window->getHeight() / 2.0f);

	m_mainWindow = CEGUI::WindowManager::getSingletonPtr()->createWindow("DefaultWindow", "Root");
	m_mainWindow->setMousePassThroughEnabled(true);
	m_guiSystem->setGUISheet(m_mainWindow);

	_CrtCheckMemory();
	return true;
}

bool UIManager::update(float deltaTime)
{
	m_keyboard->capture();
	m_mouse->capture();

	m_guiSystem->injectTimePulse(deltaTime);

    return !m_keyboard->isKeyDown(OIS::KC_ESCAPE);
}

void UIManager::setWindowExtents( int width, int height ) 
{
    // Set mouse region (if window resizes, we should alter this to reflect as well)
    const OIS::MouseState &mouseState = m_mouse->getMouseState();
    mouseState.width  = width;
    mouseState.height = height;
}

CEGUI::MouseButton UIManager::convertButton(OIS::MouseButtonID buttonID)
{
    switch (buttonID)
    {
    case OIS::MB_Left:
        return CEGUI::LeftButton;

    case OIS::MB_Right:
        return CEGUI::RightButton;

    case OIS::MB_Middle:
        return CEGUI::MiddleButton;

    default:
        return CEGUI::LeftButton;
    }
}

bool UIManager::keyPressed( const OIS::KeyEvent &e ) 
{
	bool handled = m_guiSystem->injectKeyDown(e.key);
	handled |= m_guiSystem->injectChar(e.text);

	if(!handled)
	{
		m_hotkeyManager->pressKey(e.key);
	}

    return true;
}

bool UIManager::keyReleased( const OIS::KeyEvent &e ) 
{
	bool handled = m_guiSystem->injectKeyUp(e.key);
	handled |= m_guiSystem->injectChar(e.text);

	if(!handled)
	{
		m_hotkeyManager->releaseKey(e.key);
	}

    return true;
}

bool UIManager::mouseMoved( const OIS::MouseEvent &e ) 
{
	m_guiSystem->injectMouseMove(e.state.X.rel,e.state.Y.rel);

	m_hotkeyManager->moveMouse(e.state);

	return true;
}

bool UIManager::mousePressed( const OIS::MouseEvent &e, OIS::MouseButtonID id ) 
{
	bool handled = m_guiSystem->injectMouseButtonDown(convertButton(id));

	if(!handled)
	{
		m_hotkeyManager->pressMouse(e.state, id);
	}

    return true;
}

bool UIManager::mouseReleased( const OIS::MouseEvent &e, OIS::MouseButtonID id ) 
{
	bool handled = m_guiSystem->injectMouseButtonUp(convertButton(id));

	if(!handled)
	{
		m_hotkeyManager->releaseMouse(e.state, id);
	}

    return true;
}

}